#include "loader2.h"
#include <windows.h>
#include <tlhelp32.h>
#include <stdio.h>
#include <errno.h>

int WINAPI DllMain(HINSTANCE hinst, DWORD reason, LPVOID reserved)
{
	return 1;
}

__declspec(dllexport) int Initialize(loader2_param_t *param)
{
	HANDLE snapshot;
	MODULEENTRY32 mod;
	int i, ret, success = 0;

	typedef int (*type_open_osfhandle)(void *, int);
	type_open_osfhandle func_open_osfhandle;

	typedef int (*type_dup2)(int, int);
	type_dup2 func_dup2;

	typedef int (*type_close)(int);
	type_close func_close;

	snapshot = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, GetCurrentProcessId());
	if (snapshot == NULL) {
		fprintf(stderr, "CreateToolhelp32Snapshot error: %d\n", GetLastError());
		return -1;
	}

	ZeroMemory(&mod, sizeof(mod));
	mod.dwSize = sizeof(mod);
	if (Module32First(snapshot, &mod) == 0) {
		fprintf(stderr, "Module32First error: %d\n", GetLastError());
		return -1;
	}

	do {
		func_open_osfhandle = (type_open_osfhandle)GetProcAddress(mod.hModule, "_open_osfhandle");
		if (func_open_osfhandle == NULL) continue;

		func_dup2 = (type_dup2)GetProcAddress(mod.hModule, "_dup2");
		if (func_dup2 == NULL) continue;

		func_close = (type_close)GetProcAddress(mod.hModule, "_close");
		if (func_close == NULL) continue;

		success++; /* we found all required functions in this module */

		for (i = 0; i < param->fdn; i++) {
			int fdorig;

			if (param->fds[i].fd < 3) continue;

			fdorig = func_open_osfhandle(param->fds[i].handle, param->fds[i].mode);
			if (fdorig == -1) {
				fprintf(stderr, "_open_osfhandle error (%d): %d\n", param->fds[i].fd, errno);
				success--;
				break;
			}

			if (fdorig != param->fds[i].fd) {
				if (func_dup2(fdorig, param->fds[i].fd) == -1) {
					fprintf(stderr, "_dup2 error (%d): %d\n", param->fds[i].fd, errno);
					success--;
					break;
				}
				if (func_close(fdorig) == -1) {
					fprintf(stderr, "_close error (%d): %d\n", param->fds[i].fd, errno);
					success--;
					break;
				}
			}
		}
	} while(Module32Next(snapshot, &mod));

	if (!success) {
		fprintf(stderr, "error: no good libc found\n");
		return -1;
	}

	return 0;
}
