# this plugin script wraps locdump


# $1: ind fn
# $2: c source fn
# $3: -L
# $4: highlight open tag
# $5: highlight close tag
locdump_() {
	$LIBDIR/locdump -i "$1" -c "$2" "$3" "-O$4" "-C$5"
}

locdump_html()
{
echo '<html>
<body>
<pre>'

locdump_ "$1" "$2" "$3" "<b><font color=\"red\">" "</font></b>"  | sed "s/<anon/\&lt;anon/g"

echo '
</pre>
</body>
</html>
'
}

locdump_txt()
{
	locdump_ "$1" "$2" "$3" "\`" "\`"
}


cmd_locdump()
{
	local opts loc fmt fn

	fmt=txt
	loc="-L%o:%O"

	while test $# -gt 0
	do
		case "$1" in
			-L) loc="$1" ;;
			--html) fmt=html ;;
			-*) opts="$opts $1" ;;
			*) fn="$1" ;;
		esac
		shift 1
	done

	cmd_treecc $opts "-L%f:%l [%o:%O]" -Tind "$fn" > "$fn.ind"
#	 || return
	case "$fmt" in
		html) locdump_html "$fn.ind" "$fn" "$loc" ;;
		txt) locdump_txt "$fn.ind" "$fn" "$loc" ;;
		*) echo "Unknown format $fmt" >&2 ;;
	esac
}

help_locdump()
{
	echo "Dump the tree of a C source in the \"ind\" format in the left column"
	echo "and copy the relevant parts of the C source in the right column"
	echo "highlighting the location referenced by the node of the correspinding."
	echo "\"ind\" line."
	echo ""
	echo "Usage:"
	echo "c99tree locdump [--html] [-Lloc] [treecc_opts] src"
	echo "  --html         print a colorful html instead of text"
	echo "  -Lloc          location format on the output; see the help for treecc"
	echo "  treecc_opts    any other option treecc accepts"
	echo "  src            the path to the C source to print"
	echo ""
}
