BEGIN {
	c99tree_unknown(TREE)
	map_data(TREE, "", DATA, NAMES)
	groups = "union, struct, enum"
	ws="[ \t]*"
	gsub("[\\][t]", "\t", indent_pre)
	gsub("[\\][s]", " ", indent_pre)
	gsub("[\\][t]", "\t", indent_post)
	gsub("[\\][s]", " ", indent_post)
	gsub("[\\][e]", "", indent_pre)
	gsub("[\\][e]", "", indent_post)
}

function error(msg)
{
	print "Error: " msg " (in line " NR ")" > "/dev/stderr"
}

function is_group(type)
{
	return (type == "union") || (type == "struct") || (type == "enum")
}

function loc_group_range(R, gname, UIDS  ,n,len)
{
	R["last_offs"] = ""
	R["last_path"] = ""
	R["last_uid"] = ""
	R["first_offs"] = ""
	R["first_path"] = ""
	R["first_uid"] = ""

	for(n in NAMES) {
		if (n ~ "^" gname "[.]") {
			begin=TREE[NAMES[n], C99F_LOCSTR]
			end=begin
			sub(":.*", "", begin)
			sub(".*:", "", end)
			if (end > R["last_offs"]) {
				R["last_offs"] = end
				R["last_path"] = n
				R["last_uid"] = NAMES[n]
			}
			if ((R["first_offs"] == "") || (begin < R["first_offs"])) {
				R["first_offs"] = begin
				R["first_path"] = n
				R["first_uid"] = NAMES[n]
			}
			UIDS[begin] = begin "." NAMES[n]
		}
	}
	len = asort(UIDS)
	for(n = 0; n < len; n++) {
		UIDS[n] = UIDS[n+1]
		sub("^[0-9]*[.]", "", UIDS[n])
	}
	UIDS[n+1] = ""
	UIDS["len"] = len
}

function learn_indent1(uid, IND, idx_pre, idx_post   ,tp,tk,cmt)
{
	tp = TREE[uid, C99F_TWIN_PARENT]
	cmt = 0
#	print "learn uid=" uid, tp
	for(;;) {
		tp = prv(TREE, tp)
		if (tp == "")
			break
		tk = TREE[tp, C99F_NAME]
		if ((tk == "\n") || (rk == ",") || (tk ~ "[{]$"))
			break
		if (tk ~ "[/][*].*[*][/]")
			cmt=1
		if (!cmt)
			IND[idx_pre] = TREE[tp, C99F_NAME] IND[idx_pre]
		if (tk == "/*")
			cmt=0
	}

	tp = TREE[uid, C99F_TWIN_PARENT]
	cmt = 0
	for(;;) {
		tp = nxt(TREE, tp)
		if (tp == "")
			break
		tk = TREE[tp, C99F_NAME]
		if ((tk == "\n") || (tk == "=") || (tk == ",") || (tk == "//") || (tk ~ "^[}]"))
			break
		if (tk ~ "[/][*].*[*][/]")
			cmt=1
		if (!cmt)
			IND[idx_post] = IND[idx_post] TREE[tp, C99F_NAME]
		if (tk == "*/")
			cmt=0
	}
}

function most_popular(ARR  ,n,best,bestn)
{
	best = 0
	for(n in ARR) {
		if (ARR[n] >= best) {
			bestn = n
			best = ARR[n]
		}
	}
	return bestn
}

# try to find out how indentation works
# - UIDS is an ordered list of UIDs in the group [0..n-1] where n is UIDS["len"]
# - IND is the result:
#   ["begin_pre"], ["begin_post"], ["mid_pre"], ["mid_post"], ["end_pre"], ["end_post"]
function learn_indent(UIDS, IND,   PRE,POST)
{
	learn_indent1(UIDS[0], IND, "begin_pre", "begin_post")
#	print "b |" IND["begin_pre"] "|"  IND["begin_post"] "|"
	learn_indent1(UIDS[UIDS["len"]-1], IND, "end_pre", "end_post")
#	print "e |" IND["end_pre"] "|"  IND["end_post"] "|"

	PRE[IND["begin_pre"]] = 1
	PRE[IND["post_pre"]] = 1
	POST[IND["begin_post"]] = 2
	POST[IND["post_post"]] = 2
	for(n = 1; n < UIDS["len"]-1; n++) {
		learn_indent1(UIDS[n], IND, "mid_pre", "mid_post")
		PRE[IND["mid_pre"]]+=3
		POST[IND["mid_post"]]+=3
	}
	IND["mid_pre"] = most_popular(PRE)
	IND["mid_post"] = most_popular(POST)
#	print "|" IND["mid_pre"] "|"  IND["mid_post"] "|"
}

($1 == "after") {
	if ($2 == "end-of") {
		at_type = DATA[$3, "type"]
		if (!is_group(at_type))
			error("end-of makes sense only on " groups " but " $3 " is " at_type)
		loc_group_range(R, $3, UIDS)
		at = "end"
		learn_indent(UIDS, IND)
		at_offs = R["last_offs"]+1
	}
	else if ($2 == "begin-of") {
		at_type = DATA[$3, "type"]
		if (!is_group(at_type))
			error("end-of makes sense only on " groups " but " $3 " is " at_type)
		loc_group_range(R, $3, UIDS)
		is_last = 0
		at = "begin"
		learn_indent(UIDS, IND)
		at_offs = R["first_offs"]
	}
	else {
		parent_name = $2
		if (parent_name ~ "[.]") {
			sub("[.][^.]*$", "", parent_name)
			at_type = DATA[$2, "type"]
			loc_group_range(R, parent_name, UIDS)
			learn_indent(UIDS, IND)
			at = "mid"
			at_offs=TREE[NAMES[$2], C99F_LOCSTR]
			sub(".*:", "", at_offs)
			at_offs++
		}
		else {
			parent_name = ""
			at_type = "global"
		}
	}
}

($1 == "insert") {
	val=$0
	sub("^" ws $1 ws $2 ws, "", val)
	sep1=""
	sep2=""
	ind1 = indent_pre != "" ? indent_pre : IND[at "_pre"]
	ind2 = indent_post != "" ? indent_post : IND[at "_post"]
	if ($2 == "enum-member") {
		if (at == "begin") {
			sep2=",\\n"
		}
		else if (at != "mid") {
			sep1=","
			sep2="\\n"
		}
		else
			sep1=","
		print "insert", at_offs, sep1 "\\n"  ind1 val ind2 sep2
	}
	else if (($2 == "unistr-member") || ($2 == "union-member") || ($2 == "struct-member")) {
		sep2=";"
		print "insert", at_offs,  sep1  ind1 val ind2 sep2
	}
	else
		error("unknown command " $2)
}
